@@ -66,21 +66,25 @@
 
 #include "freebcp.h"
 
+enum
+{
+	BCPFORMAT_NONE, BCPFORMAT_CHARACTER, BCPFORMAT_NATIVE, BCPFORMAT_FORMATTED
+};
+typedef int BCPFORMAT;
+
 int tdsdump_open(const char *filename);
 
-void pusage(void);
-int process_parameters(int, char **, struct pd *);
+static void pusage(void);
+static int process_parameters(int, char **, BCPPARAMDATA *);
 static int unescape(char arg[]);
-int login_to_database(struct pd *, DBPROCESS **);
+static int login_to_database(BCPPARAMDATA * pdata, DBPROCESS ** pdbproc);
 
-int file_character(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir);
-int file_native(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir);
-int file_formatted(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir);
-int setoptions (DBPROCESS * dbproc, BCPPARAMDATA * params);
-
-int err_handler(DBPROCESS * dbproc, int severity, int dberr, int oserr, char *dberrstr, char *oserrstr);
-int msg_handler(DBPROCESS * dbproc, DBINT msgno, int msgstate, int severity, char *msgtext, char *srvname, char *procname,
-		int line);
+static int setoptions(DBPROCESS * dbproc, BCPPARAMDATA * params);
+static BCPFORMAT get_format(BCPPARAMDATA * params);
+static int file_process(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir);
+static int err_handler(DBPROCESS * dbproc, int severity, int dberr, int oserr, char *dberrstr, char *oserrstr);
+static int msg_handler(DBPROCESS * dbproc TDS_UNUSED, DBINT msgno, int msgstate, int severity, char *msgtext, char *srvname,
+		       char *procname, int line);
 static int set_bcp_hints(BCPPARAMDATA *pdata, DBPROCESS *pdbproc);
 
 int
@@ -116,21 +120,27 @@ main(int argc, char **argv)
 	if (!setoptions(dbproc, &params))
 		return FALSE;
 
-	if (params.cflag) {	/* character format file */
-		ok = file_character(&params, dbproc, params.direction);
-	} else if (params.nflag) {	/* native format file    */
-		ok = file_native(&params, dbproc, params.direction);
-	} else if (params.fflag) {	/* formatted file        */
-		ok = file_formatted(&params, dbproc, params.direction);
-	} else {
-		ok = FALSE;
-	}
+	ok = file_process(&params, dbproc, params.direction);
 
 	exit((ok == TRUE) ? EXIT_SUCCESS : EXIT_FAILURE);
 
 	return 0;
 }
 
+static BCPFORMAT
+get_format(BCPPARAMDATA *params)
+{
+	if (params->cflag)
+		return BCPFORMAT_CHARACTER;
+
+	if (params->nflag)
+		return BCPFORMAT_NATIVE;
+
+	if (params->fflag)
+		return BCPFORMAT_FORMATTED;
+
+	return BCPFORMAT_NONE;
+}
 
 static int unescape(char arg[])
 {
@@ -168,7 +178,7 @@ static int unescape(char arg[])
 	return strchr(p, 0) - arg;
 }
 
-int
+static int
 process_parameters(int argc, char **argv, BCPPARAMDATA *pdata)
 {
 	extern char *optarg;
@@ -386,8 +396,31 @@ process_parameters(int argc, char **argv, BCPPARAMDATA *pdata)
 
 }
 
-int
-login_to_database(BCPPARAMDATA * pdata, DBPROCESS ** pdbproc)
+static int
+process_Eflag(BCPPARAMDATA *pdata, DBPROCESS *dbproc)
+{
+	if (pdata->Eflag) {
+
+		bcp_control(dbproc, BCPKEEPIDENTITY, 1);
+
+		if (dbfcmd(dbproc, "set identity_insert %s on", pdata->dbobject) == FAIL) {
+			fprintf(stderr, "dbfcmd failed\n");
+			return FALSE;
+		}
+
+		if (dbsqlexec(dbproc) == FAIL) {
+			fprintf(stderr, "dbsqlexec failed\n");
+			return FALSE;
+		}
+
+		while (NO_MORE_RESULTS != dbresults(dbproc))
+			continue;
+	}
+	return TRUE;
+}
+
+static int
+login_to_database(BCPPARAMDATA *pdata, DBPROCESS **pdbproc)
 {
 	LOGINREC *login;
 
@@ -455,200 +488,39 @@ login_to_database(BCPPARAMDATA * pdata, DBPROCESS ** pdbproc)
 
 }
 
-int
-file_character(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir)
+static RETCODE
+format_column(BCPPARAMDATA *pdata, DBPROCESS *dbproc, BCPFORMAT file_format, int i, int is_last_column)
 {
-	DBINT li_rowsread = 0;
-	int i;
-	int li_numcols = 0;
-	RETCODE ret_code = 0;
+	int li_coltype = SYBCHAR;
+	int host_prefixlen = 0;
+	char *host_term = NULL;
+	int host_termlen = -1;
 
-	if (FAIL == bcp_init(dbproc, pdata->dbobject, pdata->hostfilename, pdata->errorfile, dir))
-		return FALSE;
-
-	if (!set_bcp_hints(pdata, dbproc))
-		return FALSE;
-
-	if (pdata->Eflag) {
-
-		bcp_control(dbproc, BCPKEEPIDENTITY, 1);
-
-		if (dbfcmd(dbproc, "set identity_insert %s on", pdata->dbobject) == FAIL) {
-			fprintf(stderr, "dbfcmd failed\n");
-			return FALSE;
-		}
-
-		if (dbsqlexec(dbproc) == FAIL) {
-			fprintf(stderr, "dbsqlexec failed\n");
-			return FALSE;
-		}
-
-		while (NO_MORE_RESULTS != dbresults(dbproc))
-			continue;
-	}
-
-	bcp_control(dbproc, BCPFIRST, pdata->firstrow);
-	bcp_control(dbproc, BCPLAST, pdata->lastrow);
-	bcp_control(dbproc, BCPMAXERRS, pdata->maxerrors);
-
-	if (dir == DB_QUERYOUT) {
-		if (dbfcmd(dbproc, "SET FMTONLY ON %s SET FMTONLY OFF", pdata->dbobject) == FAIL) {
-			fprintf(stderr, "dbfcmd failed\n");
-			return FALSE;
-		}
-	} else {
-		if (dbfcmd(dbproc, "SET FMTONLY ON select * from %s SET FMTONLY OFF", pdata->dbobject) == FAIL) {
-			fprintf(stderr, "dbfcmd failed\n");
-			return FALSE;
-		}
-	}
-
-	if (dbsqlexec(dbproc) == FAIL) {
-		fprintf(stderr, "dbsqlexec failed\n");
-		return FALSE;
-	}
-
-	while (NO_MORE_RESULTS != (ret_code = dbresults(dbproc))) {
-		if (ret_code == SUCCEED && li_numcols == 0) {
-			li_numcols = dbnumcols(dbproc);
-		}
-	}
-
-	if (0 == li_numcols) {
-		fprintf(stderr, "Error in dbnumcols\n");
-		return FALSE;
-	}
-
-	if (bcp_columns(dbproc, li_numcols) == FAIL) {
-		fprintf(stderr, "Error in bcp_columns.\n");
-		return FALSE;
-	}
-
-	for (i = 1; i < li_numcols; ++i) {
-		if (bcp_colfmt(dbproc, i, SYBCHAR, 0, -1, (const BYTE *) pdata->fieldterm,
-			       pdata->fieldtermlen, i) == FAIL) {
-			fprintf(stderr, "Error in bcp_colfmt col %d\n", i);
-			return FALSE;
-		}
-	}
-
-	if (bcp_colfmt(dbproc, li_numcols, SYBCHAR, 0, -1, (const BYTE *) pdata->rowterm,
-		       pdata->rowtermlen, li_numcols) == FAIL) {
-		fprintf(stderr, "Error in bcp_colfmt col %d\n", li_numcols);
-		return FALSE;
-	}
-
-	bcp_control(dbproc, BCPBATCH, pdata->batchsize);
-
-	printf("\nStarting copy...\n");
-
-	if (FAIL == bcp_exec(dbproc, &li_rowsread)) {
-		fprintf(stderr, "bcp copy %s failed\n", (dir == DB_IN) ? "in" : "out");
-		return FALSE;
-	}
-
-	printf("%d rows copied.\n", li_rowsread);
-
-	return TRUE;
-}
-
-int
-file_native(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir)
-{
-	DBINT li_rowsread = 0;
-	int i;
-	int li_numcols = 0;
-	int li_coltype;
-	RETCODE ret_code = 0;
-
-	if (FAIL == bcp_init(dbproc, pdata->dbobject, pdata->hostfilename, pdata->errorfile, dir))
-		return FALSE;
-
-	if (!set_bcp_hints(pdata, dbproc))
-		return FALSE;
-
-	if (pdata->Eflag) {
-
-		bcp_control(dbproc, BCPKEEPIDENTITY, 1);
-
-		if (dbfcmd(dbproc, "set identity_insert %s on", pdata->dbobject) == FAIL) {
-			fprintf(stderr, "dbfcmd failed\n");
-			return FALSE;
-		}
-
-		if (dbsqlexec(dbproc) == FAIL) {
-			fprintf(stderr, "dbsqlexec failed\n");
-			return FALSE;
-		}
-
-		while (NO_MORE_RESULTS != dbresults(dbproc))
-			continue;
-	}
-
-	bcp_control(dbproc, BCPFIRST, pdata->firstrow);
-	bcp_control(dbproc, BCPLAST, pdata->lastrow);
-	bcp_control(dbproc, BCPMAXERRS, pdata->maxerrors);
-
-	if (dir == DB_QUERYOUT) {
-		if (dbfcmd(dbproc, "SET FMTONLY ON %s SET FMTONLY OFF", pdata->dbobject) == FAIL) {
-			fprintf(stderr, "dbfcmd failed\n");
-			return FALSE;
-		}
-	} else {
-		if (dbfcmd(dbproc, "SET FMTONLY ON select * from %s SET FMTONLY OFF", pdata->dbobject) == FAIL) {
-			fprintf(stderr, "dbfcmd failed\n");
-			return FALSE;
-		}
-	}
-
-	if (dbsqlexec(dbproc) == FAIL) {
-		fprintf(stderr, "dbsqlexec failed\n");
-		return FALSE;
-	}
-
-	while (NO_MORE_RESULTS != (ret_code = dbresults(dbproc))) {
-		if (ret_code == SUCCEED && li_numcols == 0) {
-			li_numcols = dbnumcols(dbproc);
-		}
-	}
-
-	if (0 == li_numcols) {
-		fprintf(stderr, "Error in dbnumcols\n");
-		return FALSE;
-	}
-
-	if (bcp_columns(dbproc, li_numcols) == FAIL) {
-		fprintf(stderr, "Error in bcp_columns.\n");
-		return FALSE;
-	}
-
-	for (i = 1; i <= li_numcols; i++) {
+	if (file_format == BCPFORMAT_NATIVE) {
 		li_coltype = dbcoltype(dbproc, i);
-
-		if (bcp_colfmt(dbproc, i, li_coltype, -1, -1, NULL, -1, i) == FAIL) {
-			fprintf(stderr, "Error in bcp_colfmt col %d\n", i);
-			return FALSE;
-		}
+		host_prefixlen = -1;
+	} else if (is_last_column) {
+		host_term = pdata->rowterm;
+		host_termlen = pdata->rowtermlen;
+	} else {
+		host_term = pdata->fieldterm;
+		host_termlen = pdata->fieldtermlen;
 	}
 
-	printf("\nStarting copy...\n\n");
-
-
-	if (FAIL == bcp_exec(dbproc, &li_rowsread)) {
-		fprintf(stderr, "bcp copy %s failed\n", (dir == DB_IN) ? "in" : "out");
-		return FALSE;
-	}
-
-	printf("%d rows copied.\n", li_rowsread);
-
-	return TRUE;
+	return bcp_colfmt(dbproc, i, li_coltype, host_prefixlen, -1, (const BYTE *) host_term, host_termlen, i);
 }
 
-int
-file_formatted(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir)
+static int
+file_process(BCPPARAMDATA *pdata, DBPROCESS *dbproc, DBINT dir)
 {
+	DBINT li_rowsread = 0;
+	int i;
+	int li_numcols;
 
-	int li_rowsread;
+	BCPFORMAT file_format = get_format(pdata);
+
+	if (file_format == BCPFORMAT_NONE)
+		return FALSE;
 
 	if (FAIL == bcp_init(dbproc, pdata->dbobject, pdata->hostfilename, pdata->errorfile, dir))
 		return FALSE;
@@ -656,34 +528,43 @@ file_formatted(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir)
 	if (!set_bcp_hints(pdata, dbproc))
 		return FALSE;
 
-	if (pdata->Eflag) {
-
-		bcp_control(dbproc, BCPKEEPIDENTITY, 1);
-
-		if (dbfcmd(dbproc, "set identity_insert %s on", pdata->dbobject) == FAIL) {
-			fprintf(stderr, "dbfcmd failed\n");
-			return FALSE;
-		}
-
-		if (dbsqlexec(dbproc) == FAIL) {
-			fprintf(stderr, "dbsqlexec failed\n");
-			return FALSE;
-		}
-
-		while (NO_MORE_RESULTS != dbresults(dbproc))
-			continue;
-	}
-
 	bcp_control(dbproc, BCPFIRST, pdata->firstrow);
 	bcp_control(dbproc, BCPLAST, pdata->lastrow);
 	bcp_control(dbproc, BCPMAXERRS, pdata->maxerrors);
 
-	if (FAIL == bcp_readfmt(dbproc, pdata->formatfile))
+	switch (file_format) {
+	case BCPFORMAT_FORMATTED:
+		if (FAIL == bcp_readfmt(dbproc, pdata->formatfile))
+			return FALSE;
+		break;
+
+		/* The data file does not use TDS nullable types. It uses non-nullable
+		 * representations, and a Length prefix if the target column was nullable.
+		 * The length prefix typically takes value 0 or -1 to indicate a null.
+		 * So, we need to check format of all columns.
+		 */
+	case BCPFORMAT_NATIVE:
+	case BCPFORMAT_CHARACTER:
+		li_numcols = bcp_gethostcolcount(dbproc);
+		for (i = 1; i <= li_numcols; ++i) {
+			if (format_column(pdata, dbproc, file_format, i, i == li_numcols) == FAIL) {
+				fprintf(stderr, "Error in bcp_colfmt col %d\n", i);
+				return FALSE;
+			}
+		}
+		break;
+	}
+
+	if (file_format == BCPFORMAT_CHARACTER)
+		bcp_control(dbproc, BCPBATCH, pdata->batchsize);
+
+	/* note: process_Eflag frees data needed by format_column() for NATIVE type,
+	 * so call this after the column loop. */
+	if (!process_Eflag(pdata, dbproc))
 		return FALSE;
 
 	printf("\nStarting copy...\n\n");
 
-
 	if (FAIL == bcp_exec(dbproc, &li_rowsread)) {
 		fprintf(stderr, "bcp copy %s failed\n", (dir == DB_IN) ? "in" : "out");
 		return FALSE;
@@ -694,9 +575,8 @@ file_formatted(BCPPARAMDATA * pdata, DBPROCESS * dbproc, DBINT dir)
 	return TRUE;
 }
 
-
-int
-setoptions(DBPROCESS * dbproc, BCPPARAMDATA * params)
+static int
+setoptions(DBPROCESS *dbproc, BCPPARAMDATA *params)
 {
 	RETCODE fOK;
 
@@ -770,7 +650,7 @@ set_bcp_hints(BCPPARAMDATA *pdata, DBPROCESS *pdbproc)
 	return TRUE;
 }
 
-void
+static void
 pusage(void)
 {
 	fprintf(stderr, "usage:  freebcp [[database_name.]owner.]table_name|query {in | out | queryout } datafile\n");
@@ -785,8 +665,8 @@ pusage(void)
 	fprintf(stderr, "example: freebcp testdb.dbo.inserttest in inserttest.txt -S mssql -U guest -P password -c\n");
 }
 
-int
-err_handler(DBPROCESS * dbproc, int severity, int dberr, int oserr TDS_UNUSED, char *dberrstr, char *oserrstr TDS_UNUSED)
+static int
+err_handler(DBPROCESS *dbproc, int severity, int dberr, int oserr TDS_UNUSED, char *dberrstr, char *oserrstr TDS_UNUSED)
 {
 	static int sent = 0;
 
@@ -807,9 +687,9 @@ err_handler(DBPROCESS * dbproc, int severity, int dberr, int oserr TDS_UNUSED, c
 	return INT_CANCEL;
 }
 
-int
-msg_handler(DBPROCESS * dbproc TDS_UNUSED, DBINT msgno, int msgstate, int severity,
-	    char *msgtext, char *srvname, char *procname, int line)
+static int
+msg_handler(DBPROCESS *dbproc TDS_UNUSED, DBINT msgno, int msgstate, int severity,
+ 	    char *msgtext, char *srvname, char *procname, int line)
 {
 	/*
 	 * If it's a database change message, we'll ignore it.
